from django.db import models
from django.core.serializers.json import DjangoJSONEncoder


class PublishedEventQueue(models.Model):
    exchange = models.CharField(max_length=64)
    routing_key = models.CharField(max_length=256)

    event_time = models.DateTimeField(auto_now_add=True)
    aggregate = models.CharField(max_length=64)
    event_type = models.CharField(max_length=64)
    payload = models.JSONField(encoder=DjangoJSONEncoder)




from django.core.management import BaseCommand
from services_communication.consumer.process import run_consumer
from services_communication.consumer import message_router

import logging
logging.basicConfig(level=logging.WARNING)


class Command(BaseCommand):
    help = "Run consumer"

    def handle(self, *args, **options):
        if not message_router._handlers and not message_router._default_handler:
            self.stdout.write(
                self.style.WARNING('No one handler registered in default message router. It\'s right?')
            )

        self.stdout.write(
            self.style.SUCCESS('Starting consumer')
        )

        try:
            run_consumer()
        except KeyboardInterrupt:
            self.stdout.write(self.style.SUCCESS('Consumer stopped'))





from django.core.management import BaseCommand
from django.utils import autoreload

from services_communication.publisher.process import run_publisher

import logging
logging.basicConfig(level=logging.WARNING)


class Command(BaseCommand):
    help = "Run development publisher for queued aggregate event"

    def handle(self, *args, **options):
            self.stdout.write(
                self.style.SUCCESS('Starting DEVELOPMENT publisher')
            )

            try:
                autoreload.run_with_reloader(run_publisher)
            except KeyboardInterrupt:
                self.stdout.write(self.style.SUCCESS('Publisher stopped'))





from services_communication.utils import underscoreize, camelize


json_request = lambda r: camelize(r)
json_response = lambda r: underscoreize(r.json())
file_response = lambda r: r.content
full_response = lambda r: r



import os
from collections import namedtuple
from typing import Tuple

import pika
from django.conf import settings
from django.utils.module_loading import import_string
from pika.exchange_type import ExchangeType

DEFAULT = {
    'LOG_LEVEL': 'ERROR',
    'APP_ID': None,  # Use for publisher
    'BROKER_CONNECTION_URL': 'amqp://guest:guest@localhost:5672',  # The AMQP url to connect with'
    'BROKER_CONNECTION_PARAMETERS': None,
    # 'BROKER_CONNECTION_PARAMETERS': {
    #     'host': 'localhost',
    #     'port': 5672,
    #     'virtual_host': None,
    #     'username': 'guest',
    #     'password': 'guest',
    #     'credentials': pika.PlainCredentials('guest', 'guest')
    # },
    'QUEUE': '',  # The AMQP queue for consume'
    'EXCHANGES': [
    ],  # The AMQP exchanges for publisher or consumer binds. String or tuple of name and type

    'BINDS': [
    ],  # The AMQP binds for consumer. String or tuple of exchange name and routing keys'
    'CONSUMER_CLASS': 'services_communication.broker.BlockedConsumer',
    'MESSAGE_CONSUMER': 'services_communication.consumer.message_router',
    # Callback func(basic_deliver, properties, body)

    'REST_API_HOST': 'http://localhost:8000',
    'REST_API_CREDENTIAL': None,  # {"username": "username", "password": "123456"}
    'REST_API_AUTH_URL': None  # 'api/v1/auth',
}

Exchange = namedtuple("Exchange", ['name', 'type'])
Bind = namedtuple("Bind", ['exchange', 'routing_key'])


class Settings:
    LOG_LEVEL = None
    APP_ID = None
    BROKER_CONNECTION_PARAMETERS: pika.connection.Parameters = None
    QUEUE: str = None
    EXCHANGES: Tuple[Exchange] = None
    BINDS: Tuple[Bind] = None
    MESSAGE_CONSUMER = None
    CONSUMER_CLASS = None
    REST_API_HOST = None
    REST_API_CREDENTIAL = None
    REST_API_AUTH_URL = None

    def __init__(self, default, user):
        if not user:
            user = {}

        self.BROKER_CONNECTION_PARAMETERS = self.get_broker_connection_parameters(default, user)
        self.QUEUE = self.get_value("QUEUE", default, user)
        self.MESSAGE_CONSUMER = import_string(self.get_value("MESSAGE_CONSUMER", default, user))
        self.CONSUMER_CLASS = import_string(self.get_value("CONSUMER_CLASS", default, user))

        self.APP_ID = self.get_value("APP_ID", default, user) or self.get_django_project_name()

        exchanges = self.get_value("EXCHANGES", default, user)
        self.EXCHANGES = tuple(map(self.build_exchange, exchanges))

        binds = self.get_value("BINDS", default, user)
        self.BINDS = tuple(map(self.build_bind, binds))

        self.QUEUE = self.get_value("QUEUE", default, user)

        self.REST_API_HOST = self.get_value("REST_API_HOST", default, user)
        self.REST_API_AUTH_URL = self.get_value("REST_API_AUTH_URL", default, user)
        self.REST_API_CREDENTIAL = self.get_rest_api_credential(default, user)

    def build_bind(self, bind_settings):
        if isinstance(bind_settings, str):
            return Bind(exchange=bind_settings, routing_key='#')
        if (isinstance(bind_settings, list) or isinstance(bind_settings, tuple)) and len(bind_settings) == 2:
            return Bind(exchange=bind_settings[0], routing_key=bind_settings[1])
        raise Exception(
            "MICROSERVICES_COMMUNICATION_SETTINGS.BINDS "
            "element mast by str with name "
            "or tuple with 2 elements: exchange and routing key"
        )

    def build_exchange(self, exchange_settings):
        if isinstance(exchange_settings, str):
            return Exchange(name=exchange_settings, type=ExchangeType.topic)
        if (isinstance(exchange_settings, list) or isinstance(exchange_settings, tuple)) and len(
                exchange_settings) == 2:
            return Exchange(name=exchange_settings[0], type=exchange_settings[1])
        raise Exception(
            "MICROSERVICES_COMMUNICATION_SETTINGS.EXCHANGES "
            "element mast by str with name "
            "or tuple with 2 elements: name and type"
        )

    def get_value(self, key, default, user):
        if key in user:
            return user[key]
        return default[key]

    def get_django_project_name(self):
        return os.environ['DJANGO_SETTINGS_MODULE'].split('.')[0]

    def get_rest_api_credential(self, default, user):
        credential = self.get_value("REST_API_CREDENTIAL", default, user)
        if credential:
            return credential

        username = user.get("REST_API_USERNAME", None)
        password = user.get("REST_API_PASSWORD", None)

        if username or password:
            return {
                "username": username,
                "password": password,
            }

        return None

    def get_broker_connection_parameters(self, default, user):
        parameters = self.get_value("BROKER_CONNECTION_PARAMETERS", default, user)
        if not parameters:
            connection_url = self.get_value("BROKER_CONNECTION_URL", default, user)
            return pika.URLParameters(connection_url) if connection_url else None

        assert isinstance(parameters, dict), 'MICROSERVICES_COMMUNICATION_SETTINGS.BROKER_CONNECTION_PARAMETERS mast be dict, not {}'.format(
            type(parameters))

        prepared_parameters = parameters.copy()
        for key, value in parameters.items():
            if not value:
                prepared_parameters.pop(key)

        if 'credentials' not in prepared_parameters:
            username = prepared_parameters.pop('username', 'guest')
            password = prepared_parameters.pop('password', 'guest')
            prepared_parameters['credentials'] = pika.PlainCredentials(username, password)

        return pika.ConnectionParameters(**prepared_parameters)


communication_settings = Settings(DEFAULT, getattr(settings, 'MICROSERVICES_COMMUNICATION_SETTINGS', None))



import json
from datetime import timedelta

from django.core.serializers.json import DjangoJSONEncoder
from django.utils.timezone import now

from services_communication.publisher.utils import build_publisher_by_settings
from services_communication.utils import camelize


def send_command(service_name, arguments, command_name='', timeout=None):
    meta = {}

    if timeout:
        meta['expired_time'] = now() + timedelta(seconds=timeout)

    arguments['meta'] = meta

    build_publisher_by_settings()._publish(
        exchange=service_name,
        routing_key=command_name,
        body=json.dumps(camelize(arguments), cls=DjangoJSONEncoder)
    )



import json
import time
from datetime import timedelta

import pika
from django.utils.timezone import now

from services_communication.models import PublishedEventQueue
from services_communication.publisher.utils import build_publisher_by_settings


def run_publisher():
    _check_publisher_settings()
    publisher = _get_publisher()
    while True:
        for event in PublishedEventQueue.objects.all():
            publisher.publish(
                exchange=event.exchange,
                routing_key=event.routing_key,
                body=json.dumps({
                    'eventId': event.id,
                    'eventTime': event.event_time.isoformat(),
                    'eventType': event.event_type,
                    'aggregate': event.aggregate,
                    'payload': event.payload,

                }),
            )
            event.delete()
        time.sleep(1)


def is_publisher_work(max_queue_size=0, max_delay=15):
    created_before_time = now() - timedelta(seconds=max_delay)
    queue_size = PublishedEventQueue.objects.filter(event_time__lt=created_before_time).count()

    return queue_size <= max_queue_size


def _get_publisher():
    return build_publisher_by_settings()


def _check_publisher_settings():
    pass



from services_communication.settings import communication_settings


def run_consumer():
    _check_consumer_settings()
    consumer = _get_consumer()
    consumer.run()


def _get_consumer():
    return communication_settings.CONSUMER_CLASS(
        broker_connection_parameters=communication_settings.BROKER_CONNECTION_PARAMETERS,
        queue=communication_settings.QUEUE,
        exchanges=communication_settings.EXCHANGES,
        binds=communication_settings.BINDS,
        on_message_callback=communication_settings.MESSAGE_CONSUMER,
    )


def _check_consumer_settings():
    assert communication_settings.BROKER_CONNECTION_PARAMETERS, 'Broker connection not set!'
    assert communication_settings.QUEUE, 'Broker queue for consuming not set!'



import logging
import os

from django.conf import settings


def get_logger(name, loggers={}, max_bytes=100*1000*1000, backup_count=3):
    if name not in loggers:

        level = getattr(settings, 'MICROSERVICES_COMMUNICATION_SETTINGS', {}).get('LOG_LEVEL')
        if not level:
            level = logging.DEBUG if settings.DEBUG else logging.WARNING

        logger = logging.Logger(name)

        LOG_ROOT = "../logs"
        if not os.path.exists(LOG_ROOT):
            os.makedirs(LOG_ROOT)
        # handler = logging.handlers.TimedRotatingFileHandler(os.path.join(LOG_ROOT, '{}.log'.format(name)), when="midnight", interval=1)
        handler = logging.handlers.RotatingFileHandler(os.path.join(settings.LOG_ROOT, '{}.log'.format(name)), maxBytes=max_bytes, backupCount=backup_count)
        handler.suffix = "%Y%m%d"
        handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s: %(funcName)s - %(levelname)s - %(message)s'))
        handler.setLevel(level)

        logger.addHandler(handler)

        console_handler = logging.StreamHandler()
        console_handler.setFormatter(
            logging.Formatter('%(asctime)s - %(name)s: %(funcName)s - %(levelname)s - %(message)s'))
        console_handler.setLevel(level)

        logger.addHandler(console_handler)


        loggers[name] = logger

    return loggers[name]



from services_communication.broker import BlockedPublisher
from services_communication.settings import communication_settings


def build_publisher_by_settings():

    assert communication_settings.BROKER_CONNECTION_PARAMETERS, 'Broker connection not set!'

    return BlockedPublisher(
        app_id=communication_settings.APP_ID,
        broker_connection_parameters=communication_settings.BROKER_CONNECTION_PARAMETERS,
        exchanges=communication_settings.EXCHANGES,
    )
